"""Subprocess Failure Error Definition"""
import json


class SubprocessFailureError(Exception):
    """Subprocess Failure Error"""

    command: str
    message: str
    returncode: int

    # pylint: disable=super-init-not-called,fixme
    # TODO: address linting.
    def __init__(self, command: str, message: str, returncode: int):
        # super.__init__()
        self.command = command
        self.message = message
        self.returncode = returncode

    def __str__(self):
        message: str = json.dumps({"command": self.command, "message": self.message, "returncode": self.returncode})
        return message



""" NPM package.json Config File Backend"""

from typing import Optional

from ..contacts.dtos.backend_model import BackendModel
from ..contacts.errors.unknown_argument_error import UnknownArgumentError
from .abstract_backend import AbstractBackend


class BackendPackage(AbstractBackend):
    """
    NPM model.json Config File Backend
    This backend handles interactions with NPM's native model.json file format.

    Attributes
    ----------
    DEFAULT_CONFIG_FILE
        The default config file, default: "package.json"
    """

    DEFAULT_CONFIG_FILE: str = "package.json"

    def __init__(self, config_file: Optional[str] = None, base_path: Optional[str] = None):
        super().__init__(config_file=config_file, base_path=base_path)
        self.model = BackendModel.parse_file(self.conf.resolve().as_posix())

    def init_environment(self, arguments: list[str]) -> None:
        if len(arguments) > 0:
            raise UnknownArgumentError(command="init", message="No arguments to init are supported!")
        message: str = (
            "Use `npm init` to generate a new package.json configuration file.\n"
            "For additional details see: https://docs.npmjs.com/cli/v8/using-npm/scripts"
        )
        print(message)



"""Unknown Argument Error Definition"""


class UnknownArgumentError(Exception):
    """UnknownArgumentError"""

    command: str
    message: str

    # pylint: disable=super-init-not-called,fixme
    # TODO: address linting.
    def __init__(self, command: str, message: str):
        # super.__init__()
        self.command = command
        self.message = message

    def __str__(self):
        message: str = f"Command: {self.command}" + "\n" + self.message
        return message



""" Python Config File Backend"""

import configparser
import json
from typing import Optional, Union

from ..common.utils import init_environment_argument_parser
from ..contacts.dtos.backend_model import BackendModel
from .abstract_backend import AbstractBackend


class BackendConfig(AbstractBackend):
    """
    Python Config File Backend
    This backend handles interactions with Pythons native configparser format.

    Attributes
    ----------
    DEFAULT_CONFIG_FILE
        The default config file, default: "sacr.config"
    """

    DEFAULT_CONFIG_FILE: str = "sacr.config"

    def __init__(self, config_file: Optional[str] = None, base_path: Optional[str] = None):
        super().__init__(config_file=config_file, base_path=base_path)
        self.model = self._load_config()

    def init_environment(self, arguments: list[str]) -> None:
        force: bool = init_environment_argument_parser(arguments=arguments)

        if self.conf.exists() and force or not self.conf.exists():
            # write default config
            new_config: configparser.ConfigParser = configparser.ConfigParser(delimiters="=")
            new_config.add_section(section="scripts")
            new_config.set(section="scripts", option="hello", value='"echo hello"')

            with open(self.conf.resolve().as_posix(), mode="w", encoding="utf-8") as file:
                new_config.write(file)
            print(f"Sample config created at: {self.conf.resolve().as_posix()}")
        else:
            message: str = (
                f"A configuration file already exists at {self.conf.resolve()}. To over-ride use `--force` flag."
            )
            print(message)

    def _load_config(self) -> BackendModel:
        """
        Builds the `scripts` data from a configparser config.

        Returns
        -------
        The Backend Model DTO.
        """

        model_partial: dict = {"scripts": {}}

        # Attempt to load
        if self.conf.exists():
            if self.conf.is_file():
                # load file..
                config: configparser.ConfigParser = configparser.ConfigParser(delimiters="=")
                config.read(self.conf.resolve().as_posix())

                # Process the sections
                for section in config.sections():
                    if section not in model_partial:
                        model_partial[section] = {}
                    for (key, val) in config.items(section):
                        commands: Union[list, str] = json.loads(val)
                        if isinstance(commands, str):
                            model_partial[section][key] = [commands]
                        else:
                            model_partial[section][key] = commands

        return BackendModel.parse_obj(model_partial)



""" Backend Model (Data Storage)"""

# pylint: disable=no-name-in-module,too-few-public-methods
from pydantic import BaseModel


class BackendModel(BaseModel):
    """
    BackendModel DTO

    Attributes
    ----------
    scripts: The alias : command map
    """

    scripts: dict



""" Manager Configuration Definition"""

from ..base_model import BaseModel
from .command_parameters import CommandParameters
from .config_parameters import ConfigParameters


# pylint: disable=too-few-public-methods
class ManagerConfig(BaseModel):
    """
    ManagerConfig dto

    Attributes
    ----------
    command: Manager Command Parameters DTO
    config: Manager Config Parameters DTO
    """

    command: CommandParameters
    config: ConfigParameters



""" BaseModel Definition """

# pylint: disable=no-name-in-module,too-few-public-methods
from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    """Base Model DTO"""

    class Config:
        """Global Config"""

        arbitrary_types_allowed = True
        use_enum_values = True
        validate_assignment = True



""" Manager Config Command Parameters"""
from typing import Optional

from ..base_model import BaseModel


# pylint: disable=too-few-public-methods
class CommandParameters(BaseModel):
    """
    CommandParameters DTO

    Attributes
    ----------
    timeout: The maximum duration (in seconds) a process can run before being stopped.
    """

    timeout: Optional[int]



""" Backend Abstract Definition """

import subprocess
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from ..contacts.dtos.backend_model import BackendModel
from ..contacts.errors.subprocess_failure_error import SubprocessFailureError
from ..contacts.errors.unknown_argument_error import UnknownArgumentError
from ..contacts.errors.unknown_command_error import UnknownCommandError


class AbstractBackend(ABC):
    """
    Abstract Backend DTO

    Attributes
    ----------
    config_file
        The file specific to the backend being used.
    base_path
        The path to the config file.
    DEFAULT_CONFIG_FILE
        The default config file, defined within each super-class.
    DEFAULT_CONFIG_PATH
        The default location for the configuration file, default: Path(".")
    model
        BackendModel DTO
    """

    config_file: str
    base_path: Path
    DEFAULT_CONFIG_FILE: str
    DEFAULT_CONFIG_PATH: Path = Path(".")
    model: BackendModel

    def __init__(self, config_file: Optional[str] = None, base_path: Optional[str] = None):
        if config_file:
            self.config_file = config_file
        else:
            self.config_file = self.DEFAULT_CONFIG_FILE
        if base_path:
            self.base_path: Path = Path(base_path)
        else:
            self.base_path: Path = self.DEFAULT_CONFIG_PATH

    @property
    def conf(self) -> Path:
        """
        Class Property
        Path to backend specific config file.

        Returns
        -------
        Full path of configuration file.
        """

        return self.base_path / self.config_file

    @abstractmethod
    def init_environment(self, arguments: list[str]) -> None:
        """
        Initialize Environment

        Parameters
        ----------
        arguments: The arguments to used to build the default configuration.
        """

    @staticmethod
    def _command_executor(commands: list[str], per_command_timeout: Optional[int] = None) -> None:
        """
        Handles command execution.

        Parameters
        ----------
        commands: The list of commands to execute.
        per_command_timeout: The per-command timeout threshold.
        """

        # Command executor
        for command in commands:
            with subprocess.Popen(command, shell=True) as process:
                print(f"> {command}")
                try:
                    process.wait(timeout=per_command_timeout)
                except subprocess.TimeoutExpired as error:
                    raise SubprocessFailureError(
                        command=command, message=f"Exceeded timeout limit of {per_command_timeout}", returncode=1
                    ) from error
                if process.returncode != 0:
                    if process.stdout:
                        print(process.stdout)
                    if process.stderr:
                        print(process.stderr)
                    raise SubprocessFailureError(
                        command=command, message="command failed", returncode=process.returncode
                    )

    def run_command(self, arguments: list[str], per_command_timeout: Optional[int] = None) -> None:
        """
        Run a command

        Parameters
        ----------
        arguments: The arguments to execute.
        per_command_timeout: The per-command time out threshold.
        """

        if len(arguments) != 1:
            raise UnknownArgumentError(command="run", message="Expected exactly 1 argument to run!")
        if arguments[0] in self.model.scripts:
            AbstractBackend._command_executor(
                commands=self.model.scripts[arguments[0]], per_command_timeout=per_command_timeout
            )
        else:
            raise UnknownCommandError(f"Unknown command {arguments[0]} in [scripts]")



""" Manager Config Config Parameters """

from typing import Optional

from ...backend_type import BackendType
from ..base_model import BaseModel


# pylint: disable=too-few-public-methods
class ConfigParameters(BaseModel):
    """
    ConfigParameters DTO

    Attributes
    ----------
    type: The backendtype
    file: The backend configuration filename.
    path: The path to the backend configuration file.
    """

    type: BackendType
    file: Optional[str]
    path: Optional[str]



""" Command Runner Manger Definition """

import configparser
import logging
import sys
from pathlib import Path
from typing import Optional, Union

from .backends.backend_config import BackendConfig
from .backends.backend_factory import BackendFactory
from .backends.backend_package import BackendPackage
from .common.utils import clean, init_environment_argument_parser
from .contacts.command_type import CommandType
from .contacts.dtos.manager.manager_config import ManagerConfig
from .contacts.errors.parse_error import ParseError
from .contacts.errors.subprocess_failure_error import SubprocessFailureError
from .contacts.errors.unknown_argument_error import UnknownArgumentError
from .contacts.errors.unknown_command_error import UnknownCommandError


class Manager:
    """
    Command Runner Manger

    Attributes
    ----------
    backend
        Backend for use within this manager installation.
    settings
        Manager settings configuration.
    DEFAULT_CONFIG_FILE
        The default config file to use for manager settings, default: ".sacrrc"
    DEFAULT_CONFIG_PATH
        The default location for the manager configuration file, default: Path(".")
    DEFAULT_COMMAND_TIMEOUT
        The default per command time out threshold, default: None (no timeout)
    DEFAULT_CONFIG_TYPE
        The default backend type, default: "config"
    """

    backend: Union[BackendConfig, BackendPackage]
    settings: ManagerConfig
    config_file: Path

    # Global Defaults
    DEFAULT_CONFIG_FILE: str = ".sacrrc"
    DEFAULT_CONFIG_PATH: Path = Path(".")
    DEFAULT_COMMAND_TIMEOUT: Optional[int] = None  # In seconds
    DEFAULT_CONFIG_TYPE: str = "config"

    def __init__(self, config_file: Optional[str] = None, base_path: Optional[str] = None):
        """
        Class Constractor

        Parameters
        ----------
        config_file
            The filename of the manager config file to load.
        base_path
            The local file system directory that the config file resides within.
        """

        if config_file is None:
            config_file = self.DEFAULT_CONFIG_FILE
        if base_path is None:
            base_path = self.DEFAULT_CONFIG_PATH
        else:
            base_path = Path(base_path)
        self.config_file = base_path / config_file
        self.settings = self._load_configuration(config_file=self.config_file)
        self.backend = BackendFactory.build(
            backend_type=self.settings.config.type,
            config_file=self.settings.config.file,
            base_path=self.settings.config.path,
        )

    def _load_configuration(self, config_file: Path) -> ManagerConfig:
        """
        Load Manager Configuration

        Parameters
        ----------
        config_file
            Path to for the local file system configuration file to load.

        Returns
        -------
            A ManagerConfig DTO for the configuration file.
        """

        # load the defaults
        config_partial: dict = {
            "command": {"timeout": self.DEFAULT_COMMAND_TIMEOUT},
            "config": {"type": self.DEFAULT_CONFIG_TYPE, "file": None, "path": None},
        }

        # Attempt to load
        if config_file.exists():
            if config_file.is_file():
                # load file..
                logging.getLogger(__name__).debug("Loading config file: {%s}", config_file.resolve().as_posix())
                config: configparser.ConfigParser = configparser.ConfigParser()
                config.read(config_file.resolve().as_posix())

                # Process the sections
                for section in config.sections():
                    if section not in config_partial:
                        config_partial[section] = {}
                    for (key, val) in config.items(section):
                        config_partial[section][key] = val
            else:
                logging.getLogger(__name__).warning(
                    "[SKIPPING] Config file load - Was not a file? {%s}", config_file.resolve().as_posix()
                )
        else:
            logging.getLogger(__name__).debug(
                "[SKIPPING] Config file load - Does not exist: {%s}", config_file.resolve().as_posix()
            )
        return ManagerConfig.parse_obj(config_partial)

    def main(self) -> None:
        """Main entry point for the processing of the cli command."""

        # pylint: disable=broad-except
        try:
            self._process()
        except (UnknownCommandError, UnknownArgumentError, ParseError) as error:
            # Known Errors
            logging.getLogger(__name__).debug(str(error))
            print(str(error))
            Manager.display_generic_help()
        except SubprocessFailureError as error:
            print("---- Subprocess Failure ----")
            print(error)
            sys.exit(error.returncode)
        except Exception as error:
            # We encountered an unhandled exception.  This shouldn't happen.
            print("---- Unhandled Exception ----")
            logging.getLogger(__name__).debug(str(error))
            print(type(error))
            print(str(error))
            raise error
        # pylint: enable=broad-except

    def _process(self) -> None:
        """Process CLI Arguments"""

        argument_count: int = len(sys.argv)

        if argument_count == 1:
            # CLI was run without any arguments
            Manager.display_generic_help()
        else:
            try:
                # Get the subcommand to run and execute it.
                command: CommandType = CommandType(sys.argv[1])
                self._subcommand(subcommand=command, arguments=sys.argv[2:])
            except ValueError as error:
                raise UnknownCommandError(f"Unknown command {sys.argv[1]}") from error

    def _subcommand(self, subcommand: CommandType, arguments: list[str]) -> None:
        """
        Process the given subcommand

        Parameters
        ----------
        subcommand: The subcommand to execute
        arguments: The CLI arguments for the target command.
        """

        if subcommand == CommandType.HELP:
            Manager.display_full_help()
        elif subcommand == CommandType.INIT:
            self._init_environment(arguments=arguments)
            self.backend.init_environment(arguments=arguments)
        elif subcommand == CommandType.RUN:
            self.backend.run_command(arguments=arguments, per_command_timeout=self.settings.command.timeout)
        elif subcommand == CommandType.CLEAN:
            clean(arguments=arguments)
        else:
            raise UnknownCommandError(f"Unknown command {subcommand}")

    @staticmethod
    def display_generic_help() -> None:
        """Print out summary help"""
        summary: str = "Usage: sacr <command>\n" "\n" "where <command> is one of:\n" "help, init, run, clean"

        print(summary)

    @staticmethod
    def display_full_help():
        """Print our full help"""
        summary: str = (
            "Usage: sacr <command>\n"
            "\n"
            "where <command> is one of:\n"
            "help, init, run, clean\n"
            "\n"
            "help - Displays this help dialog.\n"
            "init - Will create initial configuration file.\n"
            "   This will create default .racrrc and racr.config files in the current working directory\n"
            "run <subcommand> - Execute the defined subcommand.\n"
            "   Subcommands must be defined within a supported file (racr.config, package.json)\n"
            "clean - Perform unix like `rm -rf` like removal.\n"
        )

        print(summary)

    def _init_environment(self, arguments: list[str]) -> None:
        force: bool = init_environment_argument_parser(arguments=arguments)

        if self.config_file.exists() and force or not self.config_file.exists():
            # write default config
            new_config: configparser.ConfigParser = configparser.ConfigParser(delimiters="=")
            new_config.add_section(section="command")
            new_config.set(section="command", option="timeout", value="60")
            new_config.add_section(section="config")
            new_config.set(section="config", option="type", value="config")

            with open(self.config_file.resolve().as_posix(), mode="w", encoding="utf-8") as file:
                new_config.write(file)
            print(f"Sample config created at: {self.config_file.resolve().as_posix()}")
        else:
            message: str = (
                f"A configuration file already exists at {self.config_file.resolve()}. "
                "To over-ride use `--force` flag."
            )
            print(message)



""" Helpers """

import shutil
from pathlib import Path

from ..contacts.errors.unknown_argument_error import UnknownArgumentError


def clean(arguments: list[str]) -> None:
    """
    Removes files, folders specified.

    Parameters
    ----------
    arguments: list of files and/or folders to remove.
    """

    for item in arguments:
        item_obj: Path = Path(item)
        if item_obj.exists():
            if item_obj.is_file():
                item_obj.unlink(missing_ok=True)
            else:
                # process as dir
                shutil.rmtree(path=item_obj.resolve().as_posix(), ignore_errors=True)


def init_environment_argument_parser(arguments: list[str]) -> bool:
    """
    `init` subcommand argument parser.

    Parameters
    ----------
    arguments: the arguments

    Returns
    -------
    boolean for `force` the only supported argument.
    """

    force: bool = False
    if len(arguments) == 1:
        if arguments[0] == "--force":
            force = True
        else:
            raise UnknownArgumentError(command="init", message="Only one argument `--force` is supported.")
    elif len(arguments) > 1:
        raise UnknownArgumentError(command="init", message="Only one argument `--force` is supported.")
    return force



""" Backend Factory Definition"""

from typing import Optional, Union

from ..backends.backend_config import BackendConfig
from ..backends.backend_package import BackendPackage
from ..contacts.backend_type import BackendType


# pylint: disable=too-few-public-methods
class BackendFactory:
    """
    Backend Factory
    Factory class for the creation of backends.
    """

    @staticmethod
    def build(
        backend_type: BackendType, config_file: Optional[str] = None, base_path: Optional[str] = None
    ) -> Union[BackendConfig, BackendPackage]:
        """
        Our factory entry point
        Given the provided backend type and parameters will instantiate and return the result.

        Parameters
        ----------
        backend_type: The backend type to build.
        config_file: The configuration file to use when loading the backend.
        base_path: The location of the configuration file used when loading the backend.

        Returns
        -------
        The requested backend.
        """

        if backend_type == BackendType.CONFIG:
            return BackendConfig(config_file=config_file, base_path=base_path)
        return BackendPackage(config_file=config_file, base_path=base_path)
